from PySide6 import QtCore
from PySide6.QtWidgets import QWidget, QVBoxLayout, QPushButton, QScrollArea, QFrame, QHBoxLayout, QLineEdit
from famegui import models
from fameio.input.schema import AttributeSpecs

from famegui.config.style_config import REMOVE_BTN_STYLE, COLLAPSE_BTN_STYLE
from famegui.ui.fame_list_item import ListItem
from famegui.ui.fame_ui_elements import QFameBoldLabel


class CustomListWidget(QFrame):
    """Wrapper to enable the creation and management of a list of all types Input Panels."""
    row_creation_requested = QtCore.Signal(AttributeSpecs, QWidget)

    def __init__(
            self,
            widget_instance,
            attr_spec: AttributeSpecs,
            schema: models.Schema,
            attr_name: str,
            parent=None,
    ):
        super(CustomListWidget, self).__init__(parent)

        self.widget_instance = widget_instance
        self.attr_spec = attr_spec
        self.attr_name = attr_name
        self.schema: models.Schema = schema
        self.main_layout = QVBoxLayout(self)

        self.init_ui()

    def init_ui(self):
        self.scroll_area = QScrollArea(self)
        self.scroll_area.setWidgetResizable(True)
        self.scroll_content = QWidget(self.scroll_area)
        self.scroll_layout = QVBoxLayout(self.scroll_content)
        self.scroll_area.setWidget(self.scroll_content)

        # Add button to add new widgets
        self.add_button = QPushButton("Add", self)
        self.add_button.setFixedHeight(20)
        self.add_button.clicked.connect(self.add_widget_instance)

        # Add the button to the scroll layout
        self.scroll_layout.addWidget(QFameBoldLabel("LIST::"))
        self.scroll_layout.addWidget(self.add_button)

        # Add the scroll area to the main layout
        self.main_layout.addWidget(self.scroll_area)

        self.add_widget_instance()
        self.scroll_layout.addWidget(self.widget_instance)

        uniq_obj_name = f"customListWidgetNx-{id(self.scroll_content)}"

        self.scroll_content.setObjectName(uniq_obj_name)

        color_str = "rgba(134, 227, 206, 0.5)"

        stylesheet = f"#{uniq_obj_name} {{ border: 2px dashed grey; background-color: {color_str};  }}"

        self.scroll_content.setStyleSheet(stylesheet)
        self.setStyleSheet(stylesheet)

        self.setFixedHeight(int(self.height() * 1.2))

    def set_style(self, widget, style):
        """Sets the style of the given widget.

        Args:
          widget: The widget to set the style of.
          style: A dictionary of CSS properties and values.
        """

        for property, value in style.items():
            widget.setStyleSheet(f"{property}: {value};")

    def set_widget(self, widget_instance):
        card_wrapper_widget = ListItem(self.attr_name)
        card_wrapper_layout = card_wrapper_widget.layout()
        card_header_layout = QHBoxLayout()
        card_header_layout.addWidget(
            QFameBoldLabel(f"Block Nr.{self.scroll_layout.count() - 1}")
        )

        remove_btn = QPushButton("Remove", self)
        remove_btn.setStyleSheet(
            REMOVE_BTN_STYLE
        )
        collapse_btn = QPushButton("Collapse", self)

        collapse_btn.setStyleSheet(
            COLLAPSE_BTN_STYLE
        )

        remove_btn.setFixedHeight(30)
        remove_btn.clicked.connect(
            lambda: self.remove_widget_instance(card_wrapper_widget)
        )

        card_header_layout.addWidget(remove_btn)
        card_header = QWidget()
        card_header.setLayout(card_header_layout)
        collapse_btn.setFixedHeight(30)
        collapse_btn.clicked.connect(
            lambda: widget_instance.hide()
            if widget_instance.isVisible()
            else widget_instance.show()
        )

        card_wrapper_layout.addWidget(card_header)

        card_wrapper_widget.set_panel_widget(widget_instance)
        card_header_layout.addWidget(collapse_btn)

        uniq_widget = f"myUniqueWidget{id(card_wrapper_widget)}"
        card_wrapper_widget.setObjectName(uniq_widget)

        self.scroll_layout.addWidget(card_wrapper_widget)

        color_str = "rgba(190, 196, 254, 0.8)"

        card_header.setStyleSheet(
            f"background-color: {color_str}; border-radius: 8px; border: 2px dashed grey;"
        )

    def add_widget_instance(self):
        """Duplicate the initial list item and add it to the scroll layout."""

        if self.scroll_layout.count() == 0:
            card_wrapper_widget = ListItem(self.attr_name)

            new_widget = type(self.widget_instance)()
            new_widget: QWidget

            new_widget.setStyleSheet("background: rgb(255, 255, 49);")

            new_widget.setLayout(self.widget_instance.layout())

            card_wrapper_widget.set_panel_widget(new_widget)

            self.scroll_layout.addWidget(card_wrapper_widget)
            return
        self.row_creation_requested.emit(self.attr_spec, self)

    def remove_widget_instance(self, widget_instance):
        """Remove the given widget from list scroll layout."""
        widget_instance.deleteLater()
        self.scroll_layout.removeWidget(widget_instance)

    def _merge_data(self, data):
        """Merge input date recursively from nested input blocks to a single list of dicts."""

        merged_list = []

        for item in data:
            for key, value in item.items():
                if isinstance(value, list):
                    merged_dict = {}
                    for sub_item in value:
                        merged_dict.update(sub_item)
                    merged_list.append(merged_dict)
                elif isinstance(value, dict):
                    merged_list.extend(self._merge_data([value]))

        return merged_list

    def get_data(self):
        """Group and gather all the data from the input widget."""
        data = {}
        data_list = []

        widget_list = []
        for i in range(self.scroll_layout.count()):
            widget = self.scroll_layout.itemAt(i).widget()
            widget_list.append(widget)

            if widget:
                if widget.__class__.__name__ == "BlockItemWrapper":
                    data_list.append({str(self.attr_name): widget.get_data()})

                if isinstance(widget, QLineEdit):
                    data_list.append({str(self.attr_name): widget.get_data()})

                if widget.__class__.__name__ == "BlockTypeWrapper":
                    data_list.append({str(self.attr_name): widget.get_data()})
                    continue

                if widget.__class__.__name__ == "InputPanelWrapper":
                    data_list.append({str(self.attr_name): widget.get_data()})

                if isinstance(widget, ListItem):
                    widget: ListItem
                    data_list.append({str(self.attr_name): widget.get_data()})

                else:
                    pass

        data[self.attr_name] = data_list

        merged_data = self._merge_data(data_list)

        return merged_data



import re
import typing

import PySide6
from PySide6 import QtCore
from PySide6.QtCore import Qt
from PySide6.QtGui import QColor
from PySide6.QtWidgets import QTreeWidget, QTreeWidgetItem, QApplication

from famegui.config.runtime_consts import CONTRACT_SENDER_ID, CONTRACT_RECEIVER_ID, CONTRACT_PRODUCT_NAME
from famegui.utils import get_product_name_from_contract
from famegui.fame_tools.string_utils import extract_product_name_from_contract_identifier, extract_id_from_contract_identifier


class ExtendedTreeWidget(QTreeWidget):
    """Tree widget that uses custom event signals/triggers to communicate with the main application"""

    all_agents_deletion_requested = QtCore.Signal(str, name="del_all_agents_of_type")
    single_agent_deletion_requested = QtCore.Signal(int, name="del_single_agent")
    single_contract_deletion_requested = QtCore.Signal(
        int, int, str, name="del_single_agent"
    )  # agent_sender_id,agent_receiver_id, product_name
    agent_deletion_requested = QtCore.Signal(int, name="agent_deletion_requested")
    on_multi_contract_selected = QtCore.Signal(list,
                                               name="on_multi_contract_selected")  # list of {ref_agent_id, target_agent_id,

    # product_name, given_contract_type}

    def __init__(self, parent: typing.Optional[PySide6.QtWidgets.QWidget] = None) -> None:
        super().__init__(parent)
        self.setSelectionMode(QTreeWidget.SelectionMode.MultiSelection)
        self.itemClicked.connect(self.on_item_clicked)

        self._cross_selected_items = {}
        self._cross_selected_items_original_colors = {}


    def on_item_clicked(self, item, column):
        item.setSelected(True)
        item: QTreeWidgetItem

        modifiers = QApplication.keyboardModifiers()
        ctrl_pressed = modifiers & Qt.KeyboardModifier.ControlModifier

        if ctrl_pressed:

            self._cross_selected_items[id(item)] = item

            self._cross_selected_items_original_colors[id(item)] = item.background(0).color()
            item.setBackground(0, QColor("#FFDF64"))

    def keyReleaseEvent(self, event):

        if event.key() == Qt.Key.Key_Control:

            selected_items = [value for _, value in self._cross_selected_items.items()]

            for item_key, item in self._cross_selected_items.items():
                item.setBackground(0, self._cross_selected_items_original_colors[item_key])

            tree_item_labels = [item.text(0) for item in selected_items]

            if len(selected_items) < 2:
                self._cross_selected_items.clear()
                return

            for item in tree_item_labels:
                if not item.__contains__("#") or not item.__contains__("("):
                    self._cross_selected_items.clear()
                    super().keyReleaseEvent(event)
                    return

            selected_contracts = []

            for item in selected_items:

                parent = item.parent()
                if parent is None:
                    self._cross_selected_items.clear()
                    super().keyReleaseEvent(event)
                    return
                tree_item_label = parent.text(0)
                ref_agent_id = extract_id_from_contract_identifier(parent.text(0))

                given_contract_type = item.data(1, QtCore.Qt.ItemDataRole.UserRole)
                target_agent_id = extract_id_from_contract_identifier(item.text(0))

                product_name = extract_product_name_from_contract_identifier(item.text(0))

                sender_id, receiver_id = None, None

                if given_contract_type == "sender":
                    sender_id = ref_agent_id
                    receiver_id = target_agent_id
                else:
                    sender_id = target_agent_id
                    receiver_id = ref_agent_id

                selected_contract_item = {
                    CONTRACT_SENDER_ID: sender_id,
                    CONTRACT_RECEIVER_ID: receiver_id,
                    CONTRACT_PRODUCT_NAME: product_name
                }


                selected_contracts.append(selected_contract_item)

            self.on_multi_contract_selected.emit(selected_contracts)
            self._cross_selected_items.clear()

            super().keyReleaseEvent(event)

        super().keyReleaseEvent(event)

    def _get_id_from_item_desc(self, text: str):
        # extract id from text
        pattern = r"#(\d+)"
        match = re.search(pattern, text)

        if match:
            result = match.group(1)
            return result
        else:
            return ""

    def process_contract_deletion(
            self, text_of_selected_item: str, current_item, agent_id_ref
    ):
        selected_agent_id = current_item.parent().data(0, QtCore.Qt.UserRole)
        connection_type = current_item.data(1, QtCore.Qt.UserRole)  # sender or receiver
        operator_name_two = get_product_name_from_contract(text_of_selected_item)

        if connection_type == "sender":
            self.single_contract_deletion_requested.emit(
                int(selected_agent_id),
                int(agent_id_ref),
                operator_name_two,
            )
            return
        self.single_contract_deletion_requested.emit(
            int(agent_id_ref),
            int(selected_agent_id),
            operator_name_two,
        )

    def keyPressEvent(self, event):
        """Process key press event to process the deletion of agents and contracts"""
        if event.key() == Qt.Key_Delete:
            selected_items = self.selectedItems()

            for item in selected_items:
                parent = item.parent()
                text_of_selected_item = self.currentItem().text(0)

                if not parent:  # -> check if item has no parent to trigger the deletion of all agents of a certain type
                    self.all_agents_deletion_requested.emit(text_of_selected_item)
                    return

                agent_id_ref = self._get_id_from_item_desc(text_of_selected_item)

                if agent_id_ref:
                    # item has a parent and is a contract or single agent
                    non_agent_root_item = self.currentItem()

                    if not self.currentItem().text(0).__contains__(
                            "("):  # use display id to check if item is a contract

                        self.agent_deletion_requested.emit(int(agent_id_ref))

                        return

                    self.process_contract_deletion(
                        text_of_selected_item, non_agent_root_item, agent_id_ref
                    )

                    return



import re

from famegui.models import Contract


def get_product_name_from_contract(formatted_tree_view_contract_text: str):
    """Extract the product name from a formatted treeview contract item text"""

    pattern = r"\((.*?)\)"
    match = re.search(pattern, formatted_tree_view_contract_text)
    # Return the match if found, otherwise return None
    if match:
        return match.group(1)
    else:
        return None

def is_exactly_convertible_to_float(value: str) -> bool:
        try:
            # Convert the string to float and back to string, then compare
            return value == str(float(value))
        except ValueError:
            return False


def is_exactly_convertible_to_int(value: str) -> bool:
    try:
        # Convert the string to float and back to string, then compare
        return value == str(int(value))
    except ValueError:
        return False


def contract_fields_to_natural_types(contract_data_dict: dict, contract: Contract) -> dict:
    """Convert the contract fields to python types"""
    """ This is necessary to convert strings to ints"""
    for field_name, field_value in contract_data_dict.items():
        if contract.get_field_type(field_name) == Contract.FIELD_TYPE_INT:
            contract_data_dict[field_name] = int(field_value)
    return contract_data_dict


def get_id_from_item_desc(text: str):
    """Used to extract the id from a agent item description in the tree view"""
    pattern = r"#(\d+)"
    match = re.search(pattern, text)

    if match:
        result = match.group(1)
        return result
    else:
        return ""



import os

from famegui import models
import fameio.input.schema as fameio

_dir = os.path.dirname(__file__)
_schema_path = os.path.abspath(os.path.join(_dir, "../../../testing_resources/schemas/schema-example.yaml"))


def test_schema_loader():
    yaml_path = os.path.abspath(_schema_path)
    schema = models.Schema.load_yaml_file(yaml_path)

    assert len(schema.agent_types) == 25

    assert "Biogas" in schema.agent_types
    assert schema.agent_types["Biogas"].name == "Biogas"
    assert len(schema.agent_types["Biogas"].attributes) == 9

    assert "DemandTrader" in schema.agent_types
    assert "Loads" in schema.agent_types["DemandTrader"].attributes
    assert schema.agent_types["DemandTrader"].attributes["Loads"].is_mandatory is True
    assert schema.agent_types["DemandTrader"].attributes["Loads"].attr_type == fameio.AttributeType.BLOCK

    assert "VariableRenewableOperator" in schema.agent_types
    assert len(schema.agent_types["VariableRenewableOperator"].attributes) == 6

    assert (
        schema.agent_types["VariableRenewableOperator"].attributes["InstalledPowerInMW"].attr_type
        == fameio.AttributeType.TIME_SERIES
    )

    assert schema.agent_types["VariableRenewableOperator"].attributes["SupportInstrument"].values == [
        "FIT",
        "MPVAR",
        "MPFIX",
        "CFD",
        "CP",
        "FINANCIAL_CFD"
    ]



"""Light implementation of the Observer design pattern."""
import typing


class Observable:
    """Allows observers to be notified of something by subscribing to this class.

    Each Observable instance can be attached to some other instances to compose a hierarchy of
    observables where each level propagates (forwards) the notifications coming from its sublevels.

    Note: this implementation of the observer patter is NOT thread safe!"""

    def __init__(self):
        self._observers = []

    def attach(self, callback: typing.Callable) -> None:
        """Appends the given callback (lambda function) to the observers."""
        self._observers.append(callback)

    def notify(self) -> None:
        """Notifies all the observers"""
        for obs in self._observers:
            obs()

    def forward(self, other: "Observable") -> None:
        """Forwards the notifications coming from another observable as if they were triggered
        by the current instance)."""
        assert other != self
        other.attach(self.notify)



import random


def replace_first_number(data, replaced=False):
    """
    Recursively iterate through a dictionary or list and replace the first encountered
    integer or float with a random number.

    Args:
        data: Input dictionary or list
        replaced: Flag to track if replacement has occurred

    Returns:
        Tuple of (modified data structure, boolean indicating if replacement occurred)
    """
    if isinstance(data, dict):
        new_dict = {}
        for key, value in data.items():
            if replaced:
                new_dict[key] = value
                continue

            if isinstance(value, str) and not replaced:
                if value.__contains__(".") and value.__contains__(".csv"):
                    new_dict[key] = random.randint(1, 100)
                    replaced = True
                    continue

            if isinstance(value, (int, float)) and not replaced:
                new_dict[key] = random.randint(1, 100)
                replaced = True
            elif isinstance(value, (dict, list)):
                new_value, replaced = replace_first_number(value, replaced)
                new_dict[key] = new_value
            else:
                new_dict[key] = value
        return new_dict, replaced

    elif isinstance(data, list):
        new_list = []
        for item in data:
            if replaced:
                new_list.append(item)
                continue

            if isinstance(item, (int, float)) and not replaced:
                new_list.append(random.randint(1, 100))
                replaced = True
            elif isinstance(item, (dict, list)):
                new_value, replaced = replace_first_number(item, replaced)
                new_list.append(new_value)
            else:
                new_list.append(item)
        return new_list, replaced

    return data, replaced



import os
import logging
import typing


class AppWorkingDir:
    """Class used to find and access the files from the FAME GUI working directory"""

    _SCHEMAS_DIR_NAME = "schemas"
    _SCENARIOS_DIR_NAME = "scenarios"
    _TIMESERIES_DIR_NAME = "timeseries"
    _PROTOBUF_DIR_NAME = "protobuf"

    def __init__(self, root_dir: str):
        self._root_dir = os.path.abspath(root_dir).replace("\\", "/")

    @property
    def root_dir(self) -> str:
        return self._root_dir

    @property
    def schemas_dir(self):
        return os.path.join(self._root_dir, self._SCHEMAS_DIR_NAME)

    @property
    def scenarios_dir(self):
        return os.path.join(self._root_dir, self._SCENARIOS_DIR_NAME)

    @property
    def timeseries_dir(self):
        return os.path.join(self._root_dir, self._TIMESERIES_DIR_NAME)

    @property
    def protobuf_dir(self):
        return os.path.join(self._root_dir, self._PROTOBUF_DIR_NAME)

    def make_relative_path(self, file_path: str) -> str:
        result = os.path.abspath(file_path).replace("\\", "/")
        if result.startswith(self.root_dir):
            result = "." + result[len(self.root_dir) :]
        return result

    def make_full_path(self, path: str) -> str:
        if os.path.isabs(path):
            return path.replace("\\", "/")
        result = os.path.join(self._root_dir, path)
        return os.path.abspath(result).replace("\\", "/")

    def find_existing_child_file(self, path: str) -> typing.Optional[str]:
        abs_path = self.make_full_path(path)
        if os.path.isfile(abs_path):
            return abs_path
        logging.info(
            "failed to locate child path '{}' in working directory".format(path)
        )
        # review: Why do we need an explicit return None at this point? Could we not just remove it without
        #  any consequences? Also why do we not return a str instance as declared in the typehint? Is this
        #  a signal to the above application layer?
        return None

    def list_existing_schema_files(self) -> typing.List[str]:
        logging.debug("listing yaml files in {}".format(self.schemas_dir))
        result = []
        for filename in os.listdir(self.schemas_dir):
            if filename.endswith(".yaml"):
                # always return a relative path in UNIX style
                result.append("./{}/{}".format(self._SCHEMAS_DIR_NAME, filename))
            else:
                logging.warning(
                    "ignoring file '{}' in {}".format(filename, self.schemas_dir)
                )
        return result



import copy
import logging
from copy import deepcopy

from PySide6 import QtCore
from PySide6.QtCore import Qt
from PySide6.QtWidgets import (QWidget, QVBoxLayout,
                               QHBoxLayout, QLabel, QFrame, QScrollArea, QSizePolicy, QToolButton, QPushButton)

from famegui.ui.ui_input_panel_wrapper import MiniAttributeSpecs
from famegui_runtime_helper.attribute_dict_formatting_helper import get_full_attr_names, StringAttributeType


class SchemaWidget(QWidget):
    """Complex widget to display an attribute from root to n-depth level"""

    def __init__(self, schema_data, is_single_item,
                 attr_row_item=None, parent=None):
        super().__init__(parent)
        self.schema_data = schema_data
        self._is_single_item = is_single_item
        self._attr_row_item = attr_row_item
        self.init_ui()

        self.changed_values_tracker = {}

    def init_ui(self):
        # Create main layout
        self.main_layout = QVBoxLayout()
        self.setLayout(self.main_layout)
        self.main_layout.setAlignment(Qt.AlignmentFlag.AlignTop)

        # Add scroll area for better handling of large schemas
        scroll = QScrollArea()
        scroll.setWidgetResizable(True)
        scroll.setAlignment(Qt.AlignmentFlag.AlignTop)
        scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        scroll.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)

        # Create container widget for scroll area
        container = QWidget()
        container_layout = QVBoxLayout()
        container.setLayout(container_layout)
        container_layout.setAlignment(Qt.AlignmentFlag.AlignTop)

        # Start the recursive widget building
        self.build_widget(self.schema_data, container_layout)

        scroll.setWidget(container)

        self.main_layout.addWidget(scroll, stretch=1)

    def get_last_widget(self, contents_layout):
        """Returns the last widget from the layout."""
        if contents_layout.count() > 1:  # Exclude the plus button
            return contents_layout.itemAt(contents_layout.count() - 2).widget()
        return None

    def duplicate_last_widget(self, contents_layout):
        """Duplicates the last widget in the contents layout."""

        # Get the last widget from the layout

        last_widget = self.get_last_widget(contents_layout)

        if last_widget is not None:
            # Perform a deep copy of the last widget
            new_widget = copy.deepcopy(last_widget)
            contents_layout.insertWidget(contents_layout.count() - 1, new_widget)

    def on_list_item_added(self, contents_layout, list_schema):
        """Extend the list frame with a new item"""

        contents_layout.count()

        item_frame, item_layout = self.create_block_frame(f"Item {contents_layout.count()}", is_list_item=True,
                                                          list_item_parent_layout=contents_layout)

        list_idx = contents_layout.count()
        list_idx = list_idx - 1  # minus control panel
        list_idx = list_idx - 1  # Due too indexes starting from 0

        self.build_list_item(list_schema, contents_layout, item_frame, item_layout, list_idx,
                             f"Item {contents_layout.count()}")

    def get_list_control_panel(self, contents_layout: QVBoxLayout, list_schema: dict,
                               contents_frame: QFrame) -> QWidget:
        """
        Creates and returns a control panel widget with buttons to manage a list.

        Args:
            contents_layout (QVBoxLayout): The layout containing the list elements.
            list_schema (dict): A schema defining the structure of the list items.

        Returns:
            QWidget: The control panel widget with 'Add' and 'Remove' buttons.
        """
        # Initialize the panel
        panel = QWidget()
        panel_layout = QHBoxLayout()  # Use QHBoxLayout for buttons to be side by side
        panel.setLayout(panel_layout)

        # Configure panel appearance
        panel.setFixedHeight(40)
        panel.setStyleSheet("background-color: #9BC53D; border: 1px solid #ddd;")
        panel.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)

        # Create and style 'Add' button
        add_list_item_button = QPushButton("Add Item")
        add_list_item_button.setFixedHeight(30)
        add_list_item_button.setStyleSheet("""
            QPushButton {
                background-color: #FFFFFF;
                border: 1px solid #ddd;
                border-radius: 5px;
                padding: 5px 10px;
            }
            QPushButton:hover {
                background-color: #F0F0F0;
            }
            QPushButton:pressed {
                background-color: #E0E0E0;
            }
        """)

        add_list_item_button.clicked.connect(lambda: self.on_list_item_added(contents_layout, list_schema))

        # Create and style 'Remove' button
        remove_list_item_button = QPushButton("Remove Item")
        remove_list_item_button.setFixedHeight(30)
        remove_list_item_button.setStyleSheet("""
            QPushButton {
                background-color: #FFFFFF;
                border: 1px solid #ddd;
                border-radius: 5px;
                padding: 5px 10px;
            }
            QPushButton:hover {
                background-color: #F0F0F0;
            }
            QPushButton:pressed {
                background-color: #E0E0E0;
            }
        """)

        remove_list_item_button.clicked.connect(
            lambda: self.on_list_item_removed(contents_frame, contents_layout, list_schema))

        # Add buttons to the panel layout
        panel_layout.addWidget(add_list_item_button)
        panel_layout.addWidget(remove_list_item_button)

        return panel

    def is_at_least_one_list_item_required(self, data_dict):
        """Perform schema check if at least one list item is required to save the attribute"""

        if isinstance(data_dict["attr"], dict):
            for key, item in data_dict["attr"].items():
                if item["is_mandatory"] and not item["is_list"] == True:
                    return True

                if item["has_nested_attributes"]:
                    if self.is_at_least_one_list_item_required(item):
                        return True

        else:
            if data_dict["is_mandatory"]:
                return True
        return False

    def on_list_item_removed(self, list_item_frame, list_item_parent_layout, list_schema):
        """Remove list item by a click, detach from state( remove reference to input widget """
        to_remove_idx = list_item_parent_layout.count() - 1  # due to indexes starting from 0

        last_item = list_item_parent_layout.itemAt(to_remove_idx)  # Remove the last item
        widget = last_item.widget()  # Get the widget associated with the item

        if (to_remove_idx - 1) == 0:
            if self.is_at_least_one_list_item_required(list_schema):
                widget.setStyleSheet("background-color: #DA2C38")
                return

        if widget:
            uniq_full_names = list(set(get_full_attr_names(list_schema)))

            to_remove_idx = to_remove_idx - 1

            if to_remove_idx == -1:
                return

            self._attr_row_item.remove_state_registers(uniq_full_names, to_remove_idx)

            # list_item_parent_layout.removeWidget(widget)
            widget.setParent(None)
            # widget.deleteLater()  # Delete the widget
            del widget
        else:
            logging.warning("No widget found in the layout item")

    def create_block_frame(self, title: str, is_list_item=False, is_list=False,
                           list_schema: dict = None, is_collapsible=False, list_item_parent_layout=None):

        """Create a styled collapsible frame for blocks"""
        # Main frame
        frame = QFrame()
        frame.setFrameStyle(QFrame.Panel | QFrame.Sunken)
        frame.setStyleSheet(""" QFrame {
                background-color: #f5f5f5;
                border: 1px solid #ddd;
                border-radius: 4px;
                margin: 2px;
            } """)

        # Main layout for frame
        main_layout = QVBoxLayout()
        main_layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        frame.setLayout(main_layout)

        header_container = QWidget()
        header_container_layout = QHBoxLayout()

        # Header as a toggle button
        header_button = QToolButton()
        header_button.setText(f"➖ {title}")
        header_button.setCheckable(True)
        header_button.setChecked(True)
        header_button.setStyleSheet(""" QToolButton {
                font-weight: bold;
                color: #2c3e50;
                font-size: 14px;
                text-align: left;
                padding: 4px;
                border: none;
            }  """)

        header_button.toggled.connect(
            lambda checked: self.toggle_contents(checked, contents_frame, header_button, title))

        # Collapsible contents container
        contents_frame = QFrame()
        # contents_frame.setVisible(True)
        contents_layout = QVBoxLayout(contents_frame)

        if is_list:
            contents_layout.addWidget(self.get_list_control_panel(contents_layout, list_schema, contents_frame))

        else:
            contents_frame.setStyleSheet(""" QFrame {
                                        margin-top: 2px;
                                         } """)

        # Add header and contents to the main layout

        if is_list_item:
            left = QWidget()
            left_layout = QHBoxLayout()
            left_layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeft)
            left.setLayout(left_layout)
            left.setFixedHeight(40)
            left.setStyleSheet("background-color:#F18805")

            # Add the header_button first, before any stretch
            left_layout.addWidget(header_button)

            # Add stretch after the button to push other widgets to the right
            left_layout.addStretch(1)

            close_button = QPushButton("X")
            close_button.setStyleSheet("background-color: #FF0000; color: #FFFFFF; border: none; padding: 5px;")

            close_button.clicked.connect(lambda: self.on_list_item_removed(frame, list_item_parent_layout, list_schema))

            right = QWidget()
            right_layout = QHBoxLayout()
            right_layout.addStretch(1)
            right_layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
            right_layout.addWidget(close_button)

            right.setFixedHeight(40)
            right.setLayout(right_layout)
            right.setStyleSheet("background-color:#F0A202")
            header_container_layout.addWidget(left)
            header_container_layout.addWidget(right)
            header_container.setLayout(header_container_layout)
            main_layout.addWidget(header_container)

        else:
            main_layout.addWidget(header_button)

        main_layout.addWidget(contents_frame)

        return frame, contents_layout

    def build_list_item_control_panel(self, header_button, header_container,
                                      header_container_layout, main_layout, frame, list_item_parent_layout):
        left = QWidget()
        left_layout = QHBoxLayout()
        left_layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeft)
        left.setLayout(left_layout)
        left.setFixedHeight(40)
        # Add the header_button first, before any stretch
        left_layout.addWidget(header_button)
        # Add stretch after the button to push other widgets to the right
        left_layout.addStretch(1)
        close_button = QPushButton("X")
        close_button.clicked.connect(lambda: self.on_list_item_removed(frame, list_item_parent_layout))
        right = QWidget()
        right_layout = QHBoxLayout()
        right_layout.addStretch(1)
        right_layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
        right_layout.addWidget(close_button)
        right.setFixedHeight(40)
        right.setLayout(right_layout)
        header_container_layout.addWidget(left)
        header_container_layout.addWidget(right)
        header_container.setLayout(header_container_layout)
        main_layout.addWidget(header_container)

    def toggle_contents(self, checked, contents_frame, header_button, title):
        """Toggle visibility of the contents frame and update button text"""
        contents_frame.setVisible(checked)
        header_button.setText(f"➖ {title}" if checked else f"➕ {title}")

    def build_widget(self, data, parent_layout, parent_key="",
                     primitive_index=-1, is_primitive_list_item=False):
        """Recursively build widgets for arbitrarily deep schema data"""

        if is_primitive_list_item:
            value = data['value'] if (not isinstance(data['value'], dict)
                                      and not isinstance(data["value"], list)) else ""

            label = QLabel(f"{parent_key}:{str(value) if data is not None else 'N/A'}")

            attr_type = data["type"]

            primitive_value = data["value"] if data["value"] is not None else ""

            primitive_value = primitive_value if (not isinstance(primitive_value, list)
                                                  and not isinstance(primitive_value, dict)) else ""

            mini_attr_spec = MiniAttributeSpecs(data["attr"],
                                                attr_type,
                                                data["is_mandatory"],
                                                data["options"] if "options" in data else None)

            prim_input: QWidget = self._attr_row_item.get_primitive_input_panel(None, None,
                                                                                mini_spec=mini_attr_spec,
                                                                                preset_value=primitive_value,
                                                                                list_idx=primitive_index)

            prim_input_group_widget = QWidget()

            prim_input_group_widget.setStyleSheet("""
                background-color: #E0E0E0;
                border: 1px solid #ddd; /* Black fine border */
                border-radius: 1px;       /* Optional: Rounded corners */
                border-radius: 5px;
            """)

            prim_input_group_widget_layout = QVBoxLayout()
            prim_input_group_widget_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
            prim_input.setStyleSheet("""
                background-color: #fff;
                border: 1px solid #141414; /* Black fine border */
                border-radius: 1px;       /* Optional: Rounded corners */
                border-radius: 5px;
            """)

            prim_input_group_widget_layout.addWidget(prim_input)
            prim_input_group_widget_layout.addWidget(label)
            prim_input_group_widget.setLayout(prim_input_group_widget_layout)

            parent_layout.addWidget(prim_input_group_widget)

            return

        if isinstance(data, dict):
            # Check if the dict represents a schema node with 'type' and 'attr'
            if "type" in data and "attr" in data:

                attr_type = data["type"]

                # Process BLOCK types, which can contain nested attributes
                if attr_type == StringAttributeType.BLOCK.value:
                    # Create a frame for this block
                    is_list = data.get("is_list", False)
                    frame, frame_layout = self.create_block_frame(parent_key, is_list=is_list,
                                                                  list_schema=data)
                    # If the block is a list, handle each list item
                    if is_list:
                        if data["value"] is None:
                            item_frame, item_layout = self.create_block_frame(f"Item: {0 + 1}")

                            self.build_list_item(data, frame_layout, item_frame, item_layout, 0, parent_key)
                            parent_layout.addWidget(frame)
                            return

                        for list_idx, _ in enumerate(data["value"]):
                            item_frame, item_layout = self.create_block_frame(f"Item {list_idx + 1}")

                            self.build_list_item(data, frame_layout, item_frame, item_layout, list_idx, parent_key)

                    # If this BLOCK has attributes, process them recursively
                    elif isinstance(data["attr"], dict):
                        for key, value in data["attr"].items():
                            self.build_widget(value, frame_layout, key,
                                              primitive_index=primitive_index)

                    # Add the frame to the parent layout
                    parent_layout.addWidget(frame)

                else:
                    # Handle other attribute types (TIME_SERIES, ENUM) by displaying them directly

                    attr_value = data.get('value', 'N/A')

                    label = QLabel()

                    if isinstance(attr_value, list):
                        attr_value = attr_value[primitive_index]

                    if isinstance(attr_value, dict):
                        attr_value = ""

                    if isinstance(data['value'], list) and data["is_list"]:
                        frame, frame_layout = self.create_block_frame(parent_key, is_list=True,
                                                                      list_schema=data)
                        for list_idx, _ in enumerate(data["value"]):
                            item_frame, item_layout = self.create_block_frame(f"ItemAA {list_idx + 1}")

                            if data.get("is_list", False) and isinstance(data["value"], list):
                                copied_data = deepcopy(data)
                                picked_value = copied_data["value"][list_idx]
                                copied_data["value"] = picked_value

                                self.build_list_item(copied_data, frame_layout,
                                                     item_frame, item_layout, list_idx,
                                                     parent_key)

                                continue

                            self.build_list_item(data, frame_layout, item_frame, item_layout, list_idx,
                                                 parent_key)

                        parent_layout.addWidget(frame)
                    elif data["is_list"]:
                        frame, frame_layout = self.create_block_frame(parent_key, is_list=True,
                                                                      list_schema=data)

                        for list_idx in range(1):
                            item_frame, item_layout = self.create_block_frame(f"Item {list_idx + 1}")

                            self.build_list_item(data, frame_layout, item_frame, item_layout, list_idx,
                                                 parent_key)
                        parent_layout.addWidget(frame)

                    else:
                        from famegui.ui.attribute_row_item import AttributeTreeItem

                        self._attr_row_item: AttributeTreeItem

                        primitive_value = data["value"] if data["value"] is not None else ""

                        if isinstance(primitive_value, list):
                            primitive_value = primitive_value[primitive_index]

                        primitive_value = primitive_value if (not isinstance(primitive_value, list)
                                                              and not isinstance(primitive_value, dict)) else ""

                        mini_attr_spec = MiniAttributeSpecs(data["attr"],
                                                            attr_type,
                                                            data["is_mandatory"],
                                                            data["options"] if "options" in data else None)

                        prim_input = self._attr_row_item.get_primitive_input_panel(None, None,
                                                                                   mini_spec=mini_attr_spec,
                                                                                   preset_value=primitive_value)

                        if data["attr"] != StringAttributeType.TIME_SERIES.value:
                            parent_layout.addWidget(label)

                        parent_layout.addWidget(prim_input)

                    data_key = data["type"]

                    label_text = f"{parent_key}: {data_key}"

                    label.setText(label_text)

                    label.setWordWrap(True)

                    label.setStyleSheet("color: #34495e;")

            # Handle generic dictionaries without explicit type
            else:
                for key, value in data.items():

                    if isinstance(value, dict) or isinstance(value, list):
                        frame, frame_layout = self.create_block_frame(key)
                        self.build_widget(value, frame_layout)
                        parent_layout.addWidget(frame)
                    else:
                        # Display simple key-value pairs
                        label = QLabel(f"{key}: {value if value is not None else 'N/A'}")
                        parent_layout.addWidget(label)

        # Handle list items at any depth

        elif isinstance(data, list):
            for idx, item in enumerate(data):
                item_frame, item_layout = self.create_block_frame(f"Item: {idx + 1}")
                if isinstance(item, dict):
                    for subkey, subvalue in item.items():
                        self.build_widget(subvalue, item_layout, subkey)
                else:
                    # Directly show non-dict values in list items
                    label = QLabel(f"{idx + 1}: {item if item is not None else 'N/A'}")
                    item_layout.addWidget(label)

                parent_layout.addWidget(item_frame)

        # Display simple values at leaf nodes (like strings or numbers)
        else:
            label = QLabel(f"{parent_key}: {str(data)}  {data if data is not None else 'N/A'}")
            parent_layout.addWidget(label)

    def get_widget_index(self, layout, widget):
        """
        Get the index of a widget in a layout.

        Args:
            layout (QLayout): The layout containing the widget.
            widget (QWidget): The widget whose index we want to find.

        Returns:
            int: The index of the widget, or -1 if not found.
        """
        for i in range(layout.count()):
            if layout.itemAt(i).widget() == widget:
                return i
        return -1

    def build_list_item(self, data, frame_layout, item_frame,
                        item_layout,
                        list_idx, parent_key):

        if data["type"] != StringAttributeType.BLOCK.value:
            self.build_widget(data, frame_layout, is_primitive_list_item=True,
                              parent_key=parent_key, primitive_index=list_idx)

            return

        for idx, (key_sub, item) in enumerate(data["attr"].items()):

            if item["type"] == StringAttributeType.BLOCK.value:
                sub_data_li = data["attr"][key_sub]

                self.build_widget(sub_data_li, item_layout,
                                  key_sub, primitive_index=list_idx)

            elif isinstance(item["attr"], dict):
                for subkey, subvalue in item["attr"].items():
                    if item["attr"][subkey]["type"] == StringAttributeType.BLOCK.value:
                        self.build_widget(data["attr"][subkey], item_layout,
                                          subkey, primitive_index=list_idx)
                    else:
                        attr_value = subvalue["value"]
                        label_text = ""
                        attr_value = attr_value if attr_value is not None else "N/A"
                        attr_name = subvalue["attr"]
                        label_text += f" ={attr_name} : {attr_value}"
                        label = QLabel(label_text)
                        label.setWordWrap(True)
                        label.setStyleSheet("color: #34495e;")
                        item_layout.addWidget(label)

            else:
                # Display simple values within list items
                label = QLabel(f"{parent_key}: {item if item is not None else 'N/A'}")

                prim_attr_value = item["value"]

                if isinstance(prim_attr_value, list):

                    if list_idx < len(prim_attr_value):
                        prim_attr_value = prim_attr_value[list_idx] if list_idx != -1 else ""
                    else:
                        prim_attr_value = ""

                prim_attr_value = str(prim_attr_value) if prim_attr_value is not None else ""

                mini_attr_spec = MiniAttributeSpecs(
                    item["attr"],
                    item["type"],
                    item["is_mandatory"],
                    item["options"] if "options" in item else None
                )

                prim_input: QWidget = self._attr_row_item.get_primitive_input_panel(None, None,
                                                                                    mini_spec=mini_attr_spec,
                                                                                    preset_value=prim_attr_value,
                                                                                    list_idx=list_idx)

                display_text = item["value_key"] + f":{str(list_idx)}"

                label.setText(display_text)
                label.setWordWrap(True)
                item_layout.addWidget(label)
                prim_input = prim_input if prim_input is not None else QLabel(
                    f"Input type : {item['type']} is not supported ")

                item_layout.addWidget(prim_input)

            frame_layout.addWidget(item_frame)
